"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const DockerEvents = require("docker-events");
const Docker = require("dockerode");
const EventEmitter = require("events");
const util_1 = require("util");
/**
 * Pipeline contruido sobre docker para su utilizacion
 * en diferentes proyectos.
 */
class DockerPipeline extends EventEmitter {
    /**
     * Constructor de clase.
     * @param socketPath Direccion de, socket de docker
     */
    constructor(socketPath = '/var/run/docker.sock') {
        // Constructor de eventEmmiters
        super();
        /**
         * Bus de comunicacion interna
         * @private
         */
        this.bus = new EventEmitter();
        /**
         * Coleccion de volumenes que son creados, y seran
         * eliminados al finalizar el pipeline.
         * @private
         */
        this.collVol = new Map();
        /**
         * Coleccion de contenedores creados, y que seran
         * eliminados al finalizar el pipeline.
         * @private
         */
        this.collCont = new Map();
        // Crear la instancia de docker
        this.docker = new Docker({
            socketPath,
        });
        // Crear el manejador de eventos
        this.events = new DockerEvents({
            docker: this.docker,
        });
        // Iniciar la escucha de eventos
        this.events.start();
        // Escuchar los mensajes que suceden en docker
        this.events.on('_message', (data) => {
            // Procesar el mensaje
            this.processMessage(data);
        });
    }
    /**
     * Metodo que inicia la ejecucion de un pipeline.
     * @param PipelConf Pipeline Configuracion
     */
    start(PipelConf) {
        return __awaiter(this, void 0, void 0, function* () {
            // Variable que almacena el nombre de todos los pasos
            const steps = [];
            // Recorrer todos los pasos para almacenarlos
            for (const stp of PipelConf.steps) {
                // Almacenar
                steps.push({ Name: stp.name, Status: 'pending' });
                // Indicar su status
                this.emit('status_step', { Name: stp.name, Status: 'pending' });
            }
            // Retornar el nombre de los pasos que se ejecutaran.
            this.emit('init', { steps });
            // Ver si cuenta con definicion de volumenes
            if (!util_1.isNull(PipelConf.volumes)) {
                // Recorrer los volumenes que seran creados
                for (const vNmae of PipelConf.volumes) {
                    // Crear el volumen en una promesa
                    try {
                        yield this.docker.createVolume({ Name: vNmae });
                        // Almacenar para eliminar
                        this.collVol.set(vNmae, vNmae);
                        // Indicar que se creo el volumen
                        this.emit('create_volumen', { Name: vNmae });
                    }
                    catch (error) {
                        // Mandar eliminar todo
                        this.prune();
                        // Mandar el evento de error
                        this.emit('error_volumen', error);
                        // Indicar que finalizo el pipeline
                        this.emit('finish');
                        // throw error;
                        return; // Evitar que continue el proceso
                    }
                }
            }
            // Recorrer los pasos para ejecutarlos
            for (const stp of PipelConf.steps) {
                try {
                    // Indicar que se inicia el paso
                    this.emit('init_step', { Name: stp.name, Status: 'running' });
                    // Indicar su status
                    this.emit('status_step', { Name: stp.name, Status: 'running' });
                    // Mandar ejecutar el paso
                    yield this.executeStep(stp, PipelConf.Env || []);
                    // Indicar su status
                    this.emit('status_step', { Name: stp.name, Status: 'success' });
                    // Indicar que se finalizo el paso
                    this.emit('finish_step', { Name: stp.name, Status: 'success' });
                }
                catch (error) {
                    // Indicar su status
                    this.emit('status_step', { Name: stp.name, Status: 'failed' });
                    // Indicar que finalizo
                    this.emit('finish_step', { Name: stp.name, Status: 'failed' });
                    // Escuchar en caso de error
                    this.emit('error_step', { Error: error, Name: stp.name, Status: 'failed' });
                    // Indicar que todo finalizo
                    this.emit('finish');
                    // Mandar borrar todo
                    this.prune();
                    // throw error;
                    return;
                }
            }
            // Indicar que se termino la ejecucion del pipeline.
            this.emit('finish');
        });
    }
    /**
     * Metodo para realizar la ejecucion de un paso del pipeline
     * @param Step Paso a ejecutar
     */
    executeStep(Step, Env) {
        // Crear la variable del contenedor
        let Cont;
        // Retornar la promesa
        return new Promise((rr, ff) => __awaiter(this, void 0, void 0, function* () {
            try {
                // Crear el contenedor
                Cont = yield this.docker.createContainer({
                    Env,
                    HostConfig: {
                        Binds: Step.volumes || [],
                    },
                    Image: Step.image,
                    Tty: true,
                });
                // ALmacenar para eliminar
                this.collCont.set(Step.name, Cont.id);
            }
            catch (error) {
                // Indicar que ocurrio un error
                ff(error);
                // Evitar continuar con el proceso
                return;
            }
            try {
                // Iniciar el contenedor
                yield Cont.start();
            }
            catch (error) {
                // Emitir evento de error
                this.emit('error', error.json.message);
                // Regresar la promesa con el error
                ff(error);
                // Evitar continuar con el proceso
                return;
            }
            // Recorrer todas las tareas
            for (const cmd of Step.commands) {
                // Ejecutar el comando
                try {
                    // Ejecutar la tarea
                    yield this.executeTask(Cont, Step, cmd);
                    // Lanzar el evento
                    this.emit('task_status', {
                        Name: Step.name,
                        Status: 'success',
                        Task: cmd,
                    });
                }
                catch (error) {
                    // Lanzar el error
                    this.emit('task_message', {
                        Message: null,
                        Name: Step.name,
                        Status: 'failed',
                        Task: cmd,
                    });
                    // Lo mismo para step
                    this.emit('step_message', {
                        Message: null,
                        Name: Step.name,
                        Status: 'failed',
                        Task: cmd,
                    });
                    // Lanzar el evento
                    this.emit('task_status', {
                        Name: Step.name,
                        Status: 'failed',
                        Task: cmd,
                    });
                    // Emitir el evento de error.
                    ff(error);
                    // Evitar continuar con el pipeline
                    return;
                }
            }
            // Regresar la promesa correcta
            rr(true);
        }));
    }
    /**
     * Metodo para ejecutar task en un contenedor creado.
     * @param Container Contenedor en el que se ejecutara
     * @param Step Paso que tiene los task
     * @private
     */
    executeTask(Container, Step, Command) {
        return new Promise((rr, ff) => __awaiter(this, void 0, void 0, function* () {
            // Opciones de la tarea
            const options = {
                AttachStderr: true,
                AttachStdout: true,
                Cmd: ['/bin/bash', '-c', Command],
                Env: Step.Env || [],
            };
            // Task
            let Task = null;
            // Crear la tarea a ejecutar
            try {
                Task = yield Container.exec(options);
            }
            catch (error) {
                // Indicar que ocurrio un error
                ff(false);
                // Evitar que continue el proceso
                return;
            }
            this.bus.on(Task.id, (data) => {
                // Verificar que sea para dicho mensaje
                if (data.Action === 'exec_die') {
                    // Pasar a un entero el exitCode
                    const exitCode = parseInt(data.Actor.Attributes.exitCode, 0);
                    // Validar si fue un success
                    if (exitCode === 0) {
                        // Indicar que finalizo
                        rr(data);
                    }
                    else {
                        ff(data);
                    }
                }
                else if (data.Action === 'exec_create') {
                    this.emit('task_status', {
                        Name: Step.name,
                        Status: 'pending',
                        Task: Command,
                    });
                }
                else if (data.Action === 'exec_start') {
                    this.emit('task_status', {
                        Name: Step.name,
                        Status: 'running',
                        Task: Command,
                    });
                }
            });
            // Ejecutar la tarea
            try {
                Task.start((err, stream) => {
                    if (err) {
                        return;
                    }
                    stream.on('data', (data) => {
                        // Emitir para las tareas
                        this.emit('task_message', {
                            Message: data.toString('utf8'),
                            Name: Step.name,
                            Status: 'running',
                            Task: Command,
                        });
                        // Emitir para los step
                        this.emit('step_message', {
                            Message: data.toString('utf8'),
                            Name: Step.name,
                            Status: 'running',
                            Task: Command,
                        });
                    });
                });
            }
            catch (error) {
                // Mandar error
                this.emit('task_message', {
                    Message: error.message,
                    Name: Step.name,
                    Status: 'failed',
                    Task: Command,
                });
                // Mandar a step
                this.emit('step_message', {
                    Message: error.message,
                    Name: Step.name,
                    Status: 'failed',
                    Task: Command,
                });
                // Indicar que fallo
                ff(false);
            }
        }));
    }
    /**
     * Metodo para procesar los mensajes desde Docker
     * @param Message Mensaje enviado
     * @private
     */
    processMessage(Message) {
        if (Message.Type === 'volume') {
            // Hacer algo con la nota de volumen
        }
        else if (Message.Type === 'container') {
            // Ver que accion es la que se ejecuta
            if (Message.Action === 'create') {
                //
            }
            else if (Message.Action === 'start') {
                //
            }
            else if (Message.Action === 'die') {
                // Se elimino el contenedor
            }
            else if (Message.Action === 'exec_die') {
                // ID
                const ID = Message.Actor.Attributes.execID;
                // Avisar que se finalizo
                this.bus.emit(ID, {
                    Action: Message.Action,
                    Actor: Message.Actor,
                    Id: Message.id,
                    Status: Message.status,
                });
            }
            else if (Message.Action.split(':')[0] === 'exec_create' || Message.Action.split(':')[0] === 'exec_start') {
                // ID
                const ID = Message.Actor.Attributes.execID;
                // Notificar
                this.bus.emit(ID, {
                    Action: Message.Action.split(':')[0],
                    Actor: Message.Actor,
                    Id: Message.id,
                    Status: Message.status,
                });
            }
        }
    }
    /**
     * Metodo que realiza la eliminacion de todos los elementos
     * creados durante el pipeline.
     * @private
     * @void
     */
    prune() {
        return __awaiter(this, void 0, void 0, function* () {
            // Recorrer todos los contenedores y eliminarlos
            for (const Cont of this.collCont.entries()) {
                // Recuperar el ID del contenedor
                const IDC = Cont[1];
                try {
                    // Recuperar el Contenedor
                    const ctn = this.docker.getContainer(IDC);
                    // Detener por si acaso
                    yield ctn.stop();
                    // Eliminar
                    yield ctn.remove({ force: true });
                    // Notificar
                    this.emit('delete_container', {
                        message: `Se elimino de forma correcta el contenedor con id ${IDC}`,
                    });
                }
                catch (error) {
                    // Notificar
                    this.emit('error_delete_container', {
                        message: error.message,
                    });
                }
            }
            // Recorrer para eliminar todos los volumenes creados
            for (const volItem of this.collVol.entries()) {
                // Recuperar el nombre
                const NameV = volItem[0];
                try {
                    // Recuperar el volumen
                    const vol = this.docker.getVolume(NameV);
                    // Eliminar
                    vol.remove({ force: true });
                }
                catch (error) {
                    // Notificar
                    this.emit('error_delete_volume', {
                        message: error.message,
                    });
                }
            }
        });
    }
}
exports.DockerPipeline = DockerPipeline;
